#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
The Restriction Engine manages requests within the allowable framework of the Workspaces System requirements.
"""
import logging
import sys
from enum import Enum
from typing import Any, Callable, Generic, TypeVar

from immutable_views import DictView

import workspaces.capability.services.capability_info as info

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


T = TypeVar("T")


class Restrictions(Enum):
    CREATE_VERSION = "create_version"
    CREATE_EXECUTION = "create_execution"
    CREATE_WORKFLOW = "create_workflow"
    EDIT_PARAMS = "edit_parameters"
    EDIT_FILES = "edit_files"


class Condition(Generic[T]):
    """
    Class representing a Condition for which a Rule is valid
    """

    def __init__(self, action: Callable[[T], bool], description: str):
        self.action = action
        self.description = description

    def __call__(self, *args, **kwargs):
        return self.action(*args, **kwargs)

    def __or__(self, other):
        return Condition(lambda *args, **kwargs: self(*args, **kwargs) or other(*args, **kwargs), "or")

    def __and__(self, other):
        return Condition(lambda *args, **kwargs: self(*args, **kwargs) and other(*args, **kwargs), "and")


class ConditionFailure(Generic[T], Exception):
    """
    Class representing actions to take upon a condition failure.
     - should update the trigger object with a message to display for the user
    """

    def __init__(self, obj: T, message: str):
        self.trigger_obj = obj
        self.message = message

    def record(self):
        # TODO: figure out how to propagate message to FE
        # add failure message to trigger object in order to provide message to user
        # self.trigger_obj.json()["restriction_hit"] = self.message
        logger.info(f"{self.message}")
        return self.trigger_obj


class RequestRestrictions:
    """
    Class representing possible restrictions on request objects
    """

    is_mutable = Condition(lambda request: not request.sealed, "request is mutable")
    is_first_version = Condition(lambda request: len(request.versions) == 0, "is first version")
    no_mutable_version_exists = Condition(
        lambda request: all(version.sealed for version in request.versions),
        "no mutable version exists",
    )
    not_qa_passed = Condition(
        lambda request: (request.state == "Complete" and not request.capability.requires_qa)
        or (request.state != "Complete" and request.capability.requires_qa)
        or (request.state != "Complete" and not request.capability.requires_qa),
        "not Complete by QA Pass",
    )


class VersionRestrictions:
    """
    Class representing possible restrictions on version objects
    """

    is_mutable = Condition(lambda version: not version.sealed, "version is mutable")
    is_first_execution = Condition(lambda version: len(version.executions) == 0, "is first execution")
    no_valid_execution = Condition(
        lambda version: all(execution.state == "Error" for execution in version.executions),
        "no valid execution",
    )


class ExecutionRestrictions:
    """
    Class representing possible restrictions on execution objects
    """

    is_first_workflow = Condition(lambda execution: execution.current_workflow_request_id is None, "is first workflow")


class Effect(Generic[T]):
    """
    Class representing an action the system can take based on Condition outcome
    """

    def __init__(self, action: Callable[[Any, T], Any], description: str):
        self.action = action
        self.description = description

    def __call__(self, *args, **kwargs):
        return self.action(*args, **kwargs)


class RequestEffects:
    create_new_version = Effect(
        lambda capability_info, request, parameters: info.RestrictedInfo(capability_info.session).create_version(
            request=request, parameters=parameters
        ),
        "Create New Version",
    )


class VersionEffects:

    create_new_execution = Effect(
        lambda capability_info, version, args: info.RestrictedInfo(capability_info.session).create_execution(
            version=version
        ),
        "Create New Version",
    )
    edit_version_parameters = Effect(
        lambda capability_info, version, args: info.RestrictedInfo(capability_info.session).edit_parameters(
            version=version
        ),
        "Edit Parameters",
    )
    edit_version_files = Effect(
        lambda capability_info, version, args: info.RestrictedInfo(capability_info.session).edit_files(version=version),
        "Edit Files",
    )


class ExecutionEffects:
    create_new_workflow = Effect(
        lambda capability_info, execution, args: info.RestrictedInfo(capability_info.session).create_workflow(
            execution=execution,
            args=args,
        ),
        "Create New Workflow Request",
    )


class Rule(Generic[T]):
    """
    Class representing a Rule, containing a Condition object and and Effect object.
    """

    def __init__(self, name: str, condition: Condition[T], effect: Effect[T]):
        self.rule_name = name
        self.condition = condition
        self.effect = effect


class RestrictionEngine:
    """
    The Ruleset manager for the Workspaces Capability Restrictions
        - If the Rule Condition is valid, apply the Effect
            - All listed Conditions must pass to apply Effect
    """

    # Restriction Conditions
    request = RequestRestrictions()
    version = VersionRestrictions()
    execution = ExecutionRestrictions()

    restriction_map = DictView(
        {
            "create_version": Rule(
                "Create New Version",
                (request.is_first_version | request.no_mutable_version_exists)
                & request.not_qa_passed
                & request.is_mutable,
                RequestEffects.create_new_version,
            ),
            "create_execution": Rule(
                "Create New Execution",
                version.is_first_execution | version.no_valid_execution,
                VersionEffects.create_new_execution,
            ),
            "edit_parameters": Rule(
                "Edit Parameters",
                version.is_mutable,
                VersionEffects.edit_version_parameters,
            ),
            "edit_files": Rule(
                "Edit Files",
                version.is_mutable,
                VersionEffects.edit_version_files,
            ),
            "create_workflow": Rule(
                "Create New Workflow Request",
                execution.is_first_workflow,
                ExecutionEffects.create_new_workflow,
            ),
        }
    )

    def __init__(self, capability_info):
        self.capability_info = capability_info

    def trigger(self, trigger_obj: Any, trigger_action: Restrictions, trigger_args: Any = None) -> Any:
        restriction = self.get_restriction_map(trigger_action)

        if self.evaluate_rules(trigger_obj, restriction):
            return self.run_effects(trigger_obj, restriction, trigger_args)
        else:
            return ConditionFailure(
                trigger_obj,
                f"Restriction Hit: {restriction.rule_name} is not permitted. Condition '{restriction.condition.description}' failed",
            ).record()

    def get_restriction_map(self, trigger_action: Restrictions) -> Rule:
        return self.restriction_map.get(trigger_action.value)

    @staticmethod
    def evaluate_rules(trigger_obj: Any, restriction: Rule) -> bool:
        return restriction.condition(trigger_obj)

    def run_effects(self, trigger_obj: Any, restriction: Rule, trigger_args: Any = None):
        return restriction.effect(self.capability_info, trigger_obj, trigger_args)
